/* naGuidoo.c - adapted for gmn2midi, v1.0 */

// debug tool - uncomment next line to activate debug messages:
//#define DEBUG_ON

#include <stdio.h>
#include <math.h>

#include "naguido.hpp"
#include "naguido.h"
#include "guido.h"

#include <MidiNoteOn.hpp>
#include <MidiNoteOff.hpp>
#include <MidiMetaEvent.hpp>
#include <MidiProgramChange.hpp>
#include <MidiPitchBendChange.hpp>
#include <MidiControlChange.hpp>

// ************************************************************** //
// Globale Variables to generate a MIDI-File                      //
// ************************************************************** //
class NaGuido
{
public:
  // Midi-Class
  Midi *midi;
  MidiTrackChunk *currentMidiTrackChunk;

  // NoteOffList
  MidiMessageList midiNoteOffList;

  // Midi-Settings
  unsigned short quaterPerMinute;
  float intensity;
  int currentDurPosN, currentDurPosD;
  signed long delayAddieren;

  // current used channe
  unsigned char currentChannel;
  bool useChannel[16];

  // to calculate Tempo changes
  bool multipelTempo;
  unsigned short currentQuaterPerMinute;

  // pointer to logFile
  FileIO *logFile;

  // DynamikList for \intens
  DynamicList dynamicList;

  // Repeats
  MidiMessageList::const_iterator *repeatBeginIter;
  MidiMessageList::const_iterator *repeatEndIter;
  long repeatNumber, currentRepeatNumber;

  // Tag tie
  bool tie;
  MidiNoteOff *tieMidiNoteOff;

  // factor to multiply to midiNoteOff.Delay for tag //staccato etc
  float delayFactor;
  float factorDefault;
  float factorStaccato;
  float factorTenuto;
  float factorSlur;

  // factor to multiply for fermata
  float factorFermata;
  bool fermataActiv;

  // factor to multiply to midiNote...dynamic
  float intensityFactor;
  float factorAccent;
  float factorMarcato;

  // crescendo/decrescendo
  MidiMessageList::const_iterator *crescendoBeginIter;

  // tag \oct
  signed long oct;

  // cue notes
  bool cueActiv;

  // riterdando/accelerando
  MidiMessageList::const_iterator *riterdandoBeginIter;

  // for alter
  unsigned long currentAlter;

  // for Tremulo
  bool tremActiv;
  signed long tremPar;

  // for tag trill, modent, turn, grace
  enum TmtgActiv {normal = 0, trill = 1, mord = 2, turn = 3, grace = 4};
  TmtgActiv tmtgActiv;
  TmtgActiv oldTmtgActiv;
  signed long tmtgPar;
  signed long oldTmtgPar;
  MidiTrackChunk tmtgMidiMessageList;
  bool changeDelayAddieren;
} naGuido;

// ************************************************************** //
// Functions to control Midi                                      //
// ************************************************************** //
// get ticksPerQuater from MidiHeader
unsigned short ticksPerQuater ()
{
  // decide if multipelTempo is allowed
  if (naGuido.multipelTempo)
    return (naGuido.midi->header.division () * naGuido.quaterPerMinute /
            naGuido.currentQuaterPerMinute);
  else
    return (naGuido.midi->header.division ());
}

// write MidiNoteOff, is Position reached
void writeMidiNoteOffList (int durPosN, int durPosD, bool all = false);
void writeMidiNoteOffList (int durPosN, int durPosD, bool all)
{
  // calculate Delay of this Position and set currentDurPos to durPos
  // do not change delay if not changeDelayAddieren
  if (!all&&naGuido.changeDelayAddieren)
    naGuido.delayAddieren += ticksPerQuater () * 4 * durPosN / durPosD -
                             ticksPerQuater () * 4 * naGuido.currentDurPosN /
                             naGuido.currentDurPosD;
  naGuido.currentDurPosN = durPosN;
  naGuido.currentDurPosD = durPosD;
  // check it trill, modent, turn, grace and set changeDelayAddieren
  naGuido.changeDelayAddieren = (naGuido.tmtgActiv == NaGuido::normal);

  // look at all messages
  MidiMessageList::iterator iter;
  MidiMessageList::iterator iterLowest;

  do
    {
      iterLowest = naGuido.midiNoteOffList.end ();

      // find Message, to write first
      for (iter = naGuido.midiNoteOffList.begin ();
           iter != naGuido.midiNoteOffList.end ();
           ++iter)
        {
          if (((*iter)->delay () <= naGuido.delayAddieren)||all)
            if ((iterLowest != naGuido.midiNoteOffList.end ()) ?
                ((*iter)->delay () < (*iterLowest)->delay ()) : true)
              iterLowest = iter;
        }

      // Work on found Element
      if (iterLowest != naGuido.midiNoteOffList.end ())
        {
          // get MidiNoteOff form List and copy, then delete from list
          // and add this to midiTrackChunk
          MidiMessage *midiMessage = (*iterLowest)->copy ();
          naGuido.midiNoteOffList.erase (iterLowest);
          *(naGuido.currentMidiTrackChunk) += midiMessage;

          // reduce all Messages by lowest Delay and delayAddieren
          if (!all)
            naGuido.delayAddieren -= midiMessage->delay ();
          for (iter = naGuido.midiNoteOffList.begin ();
               iter != naGuido.midiNoteOffList.end ();
               ++iter)
             (*iter)->setDelay((*iter)->delay () - midiMessage->delay ());

          // set iterLowest to begin, and prove next message
          iterLowest = naGuido.midiNoteOffList.begin ();
        }
    }
  while ((iterLowest != naGuido.midiNoteOffList.end ()));
}

// **************************************************************** //
// read GMN and handle Midi                                         //
// **************************************************************** //
void na_parse (Midi& midi, const MString& guidoDateiName,
               unsigned short quaterPerMinute, float intensity,
               bool *useChannel, bool multipelTempo, FileIO *logFile,
               DynamicList &dynamicList, float factorDefault,
               float factorStaccato, float factorTenuto, float factorSlur,
               float factorFermata, float factorAccent, float factorMarcato)
{
  // Initialize Midi
  naGuido.midi = &midi;
  naGuido.quaterPerMinute = quaterPerMinute;
  naGuido.intensity = intensity;
  naGuido.multipelTempo = multipelTempo;
  naGuido.logFile = logFile;
  naGuido.dynamicList = dynamicList;
  naGuido.factorDefault = factorDefault;
  naGuido.factorStaccato = factorStaccato;
  naGuido.factorTenuto = factorTenuto;
  naGuido.factorSlur = factorSlur;
  naGuido.factorFermata = factorFermata;
  naGuido.factorAccent = factorAccent;
  naGuido.factorMarcato = factorMarcato;

  // set channels to use
  if (useChannel != NULL)
    memcpy (naGuido.useChannel, useChannel, sizeof (bool) * 16);
  else
    memset (naGuido.useChannel, true, sizeof (bool) * 16);

  // set MidiFormat 1 (several Tracks for one movement)
  naGuido.midi->header.setFormat (1);

  // call parser
  gd_parse (guidoDateiName);
}


// **************************************************************** //
// interface functions: seqment and sequence                        //
// called from parser actions                                       //
// **************************************************************** //
// open Segment
void na_segmInit()
{
  DEBUG(*(naGuido.logFile) << "na_segmInit()\n";)

  // create track 0 for tempo
  *(naGuido.midi) += new MidiTrackChunk ();

  // set temp-Message
  MidiMetaSetTempo *midiMetaSetTempo = new MidiMetaSetTempo ();
  // usecPerQuater = usecPerMin / qaterPerMinute
  midiMetaSetTempo->setTempo (60000000 / naGuido.quaterPerMinute);
  **(naGuido.midi->begin ()) += midiMetaSetTempo;

  // for channel 0 to 15
  for (unsigned short channel = 0; channel <= 15; ++channel)
    {
      // set pitch bend sensitivity
      MidiControlChange *midiControlChange = new MidiControlChange ();
      midiControlChange->setChannel (channel);
      midiControlChange->setControl (0x65); // Registerd Parameter Number
      midiControlChange->setValue (0);
      **(naGuido.midi->begin ()) += midiControlChange;
      midiControlChange = new MidiControlChange ();
      midiControlChange->setChannel (channel);
      midiControlChange->setControl (0x64);  // Registerd Parameter Number
      midiControlChange->setValue (0);
      **(naGuido.midi->begin ()) += midiControlChange;

      // reset pitch bend position
      MidiPitchBendChange *midiPitchBendChange = new MidiPitchBendChange ();
      midiPitchBendChange->setChannel (channel);
      midiPitchBendChange->setPitchBend (0x40 << 7); // middel position
      **(naGuido.midi->begin ()) += midiPitchBendChange;
    }

  // set uesed channel to the last one
  naGuido.currentChannel = 15;
}

// end of Seqment
void na_segmExit()
{
  DEBUG(*(naGuido.logFile) << "na_segmExit()\n";)
  // don't do anything
}

// create voice
void na_seqInit()
{
  DEBUG(*(naGuido.logFile) << "na_seqInit()\n";)

  // create MidiTrack
  naGuido.currentMidiTrackChunk = new MidiTrackChunk ();
  *(naGuido.midi) += naGuido.currentMidiTrackChunk;

  // initialize Midi-settings
  naGuido.currentDurPosN = 0;
  naGuido.currentDurPosD = 1;
  naGuido.delayAddieren = 0;
  naGuido.repeatBeginIter = 0;
  naGuido.repeatEndIter = 0;
  naGuido.repeatNumber = 0;
  naGuido.currentRepeatNumber = 0;
  naGuido.tie = false;
  naGuido.tieMidiNoteOff = 0;
  naGuido.delayFactor = naGuido.factorDefault;
  naGuido.crescendoBeginIter = 0;
  naGuido.oct = 0;
  naGuido.intensityFactor = 1.0;
  naGuido.fermataActiv = false;
  naGuido.cueActiv = false;
  naGuido.riterdandoBeginIter = 0;
  naGuido.currentQuaterPerMinute= naGuido.quaterPerMinute;
  naGuido.currentAlter = 0x40 << 7;
  naGuido.tremActiv = false;
  naGuido.tremPar = -1;
  naGuido.tmtgActiv = NaGuido::normal;
  naGuido.oldTmtgActiv = NaGuido::normal;
  naGuido.tmtgPar = -1;
  naGuido.oldTmtgPar = -1;
  naGuido.changeDelayAddieren = true;

  // add to currentChannel one. If it is > 15 set it to 0
  // check if it is allowed to use the channel
  do
    if (++(naGuido.currentChannel) > 15)
      naGuido.currentChannel = 0;
  while (!naGuido.useChannel[naGuido.currentChannel]);
}

// end of Voice
void na_seqExit()
{
  DEBUG(*(naGuido.logFile) << "na_seqExit()\n";)

  // write remaining MidiNoteOff
  writeMidiNoteOffList (0, 1, true);

  // delete Classes
  if (naGuido.repeatBeginIter != 0)
    delete (naGuido.repeatBeginIter);
  if (naGuido.repeatEndIter != 0)
    delete (naGuido.repeatEndIter);
  if (naGuido.riterdandoBeginIter != 0)
    delete (naGuido.riterdandoBeginIter);
}

// ****************************************************************** //
// append Note                                                        //
// ****************************************************************** //
void na_seqAppendNote(int pc, int oct, int acc, int durN, int durD, int dots,
                      int durPosN, int durPosD)
{
  DEBUG(*(naGuido.logFile) << "na_seqAppendNote(pc=" << (long int)pc << ",oct=" << (long int)oct << ",acc=" << (long int)acc << ", durN=" << (long int)durN << ",durD=" << (long int)durD << ",dots=" << (long int)dots << ",durPosN=" << (long int)durPosN << ",durPosD=" << (long int)durPosD << ")\n";)

  // if there are som MidiNoteOff try to write the messages
  writeMidiNoteOffList (durPosN, durPosD);

  // insert note, if ther is no rest and not cueActiv
  if ((pc!=REST)&&!naGuido.cueActiv)
    {
      // calculate pitch of note
      unsigned char key = 48 + (oct * 12 + pc) + acc;

      // and calculate midiDelay (dotted notes!)
      unsigned long midiDelay;
      for (midiDelay = ticksPerQuater () * 4 * durN / durD; dots > 0; --dots)
        midiDelay += ticksPerQuater () * 4 * durN / durD / dots;

      // write MidiNoteOn in Track
      MidiNoteOn *midiNoteOn = new MidiNoteOn;
      midiNoteOn->setDelay (naGuido.delayAddieren);
      // if trill, modent, turn, grace do not reset delayAddieren
      if (naGuido.tmtgActiv == NaGuido::normal)
	naGuido.delayAddieren = 0;
      midiNoteOn->setChannel (naGuido.currentChannel);
      midiNoteOn->setKey (key + naGuido.oct * 12);
      // check intensity
      if (127 * naGuido.intensity * naGuido.intensityFactor < 127.0)
        midiNoteOn->setDynamic ((unsigned)(127 * naGuido.intensity * naGuido.intensityFactor));
      else
        midiNoteOn->setDynamic (127);
      // if fermataActiv add some time to delay
      if (naGuido.fermataActiv)
        midiDelay += naGuido.delayAddieren = (unsigned long)
                        ((float)midiDelay * (naGuido.factorFermata - 1.0));
      // trill was activ
      if (naGuido.oldTmtgActiv == NaGuido::trill)
	{
	  // calculate trillCount check oldTmtgPar to calculate default delay
	  signed long trillCount;
	  if (naGuido.oldTmtgPar > 0)
	    trillCount = midiDelay / naGuido.tmtgMidiMessageList.length ();
	  else
	    trillCount = 3; // 8 / 2 - 1
	  // get first MidiNoteOn
	  MidiTrackChunk::const_iterator iter = naGuido.tmtgMidiMessageList.begin ();
	  MidiMessage *midiMessageTemp = (*iter)->copy ();
	  // set delay and write first note
	  midiMessageTemp->setDelay (midiNoteOn->delay ());
	  *(naGuido.currentMidiTrackChunk) += midiMessageTemp;
	  // delay for midiNoteOffList
	  unsigned long trillDelayAddieren = 0;
	  // write the other Messages
          for (++iter; iter != naGuido.tmtgMidiMessageList.end (); ++iter)
	    {
	      // get MidiMessage
	      midiMessageTemp = (*iter)->copy ();
	      // set new Delay, if MidiMoteOn dont change Delay
	      if (midiMessageTemp->id () != 0x90) // MidiNoteOn 
		// if tmtgpar was not specifed calulate default delay
		if (naGuido.oldTmtgPar > 0)
		  trillDelayAddieren += midiMessageTemp->delay ();
		else
		  trillDelayAddieren += midiDelay / 8;
	      midiMessageTemp->setDelay (trillDelayAddieren);
	      // put MidiMessage in midiNoteOffList
	      naGuido.midiNoteOffList += midiMessageTemp;
	    }
	  for (--trillCount; trillCount > 0; --trillCount)
	    {
	      // ATTENTION internal compiler of gcc if ther is no new 
	      // iterator or ther is a for-loop
	      MidiTrackChunk::const_iterator newIter = naGuido.tmtgMidiMessageList.begin ();
	      while (newIter != naGuido.tmtgMidiMessageList.end ())
		{
		  // get MidiMessage
		  midiMessageTemp = (*newIter)->copy ();
		  // set new Delay, if MidiMoteOn dont change Delay
		  if (midiMessageTemp->id () != 0x90) // MidiNoteOn 
		    // if tmtgpar was not specifed calulate default delay
		    if (naGuido.oldTmtgPar > 0)
		      trillDelayAddieren += midiMessageTemp->delay ();
		    else
		      trillDelayAddieren += midiDelay / 8;
		  midiMessageTemp->setDelay (trillDelayAddieren);
		  // put MidiMessage in midiNoteOffList
		  naGuido.midiNoteOffList += midiMessageTemp;
		  ++newIter;
		}
	    }
	  // set delay of old MidiNoteOn and write it
	  midiNoteOn->setDelay (trillDelayAddieren);
	  naGuido.midiNoteOffList += midiNoteOn;
	  // delate tmtgMidiMessageList
	  naGuido.tmtgMidiMessageList.erase (naGuido.tmtgMidiMessageList.begin (), 
					     naGuido.tmtgMidiMessageList.end ());
	}
      else if ((naGuido.oldTmtgActiv == NaGuido::mord)||
	       (naGuido.oldTmtgActiv == NaGuido::turn)||
	       (naGuido.oldTmtgActiv == NaGuido::grace))
      // mord, turn, grace were activ
	{
          // get first MidiNoteOn
	  MidiTrackChunk::const_iterator iter = naGuido.tmtgMidiMessageList.begin ();
	  MidiMessage *midiMessageTemp = (*iter)->copy ();
	  // set delay and write first note
	  midiMessageTemp->setDelay (midiNoteOn->delay ());
	  *(naGuido.currentMidiTrackChunk) += midiMessageTemp;
	  // delay for midiNoteOffList
	  unsigned long mtgDelayAddieren = 0;
	  // write the other Messages
          for (++iter; iter != naGuido.tmtgMidiMessageList.end (); ++iter)
	    {
	      // get MidiMessage
	      midiMessageTemp = (*iter)->copy ();
	      // set new Delay, if MidiMoteOn dont change Delay
              if (midiMessageTemp->id () != 0x90) // MidiNoteOn
		// if tmtgpar was not specifed and tag was not grace 
		// calulate default delay
		if ((naGuido.oldTmtgPar > 0)||
		    (naGuido.oldTmtgActiv == NaGuido::grace))
		  mtgDelayAddieren += midiMessageTemp->delay ();
		else
		  mtgDelayAddieren += midiDelay / 8;
	      midiMessageTemp->setDelay (mtgDelayAddieren);
	      // put MidiMessage in midiNoteOffList
	      naGuido.midiNoteOffList += midiMessageTemp;
	    }
	  // set delay of old MidiNoteOn and write it
	  midiNoteOn->setDelay (mtgDelayAddieren);
	  naGuido.midiNoteOffList += midiNoteOn;
	  // delate tmtgMidiMessageList
	  naGuido.tmtgMidiMessageList.erase (naGuido.tmtgMidiMessageList.begin (), 
					     naGuido.tmtgMidiMessageList.end ());
	}
      // if tieMidiNoteOff is set, add Dummy instead of midiNoteOn Message
      if (naGuido.tieMidiNoteOff == 0)
	// if trill, modent, turn, grace store MidiMessages in separat list
	if (naGuido.tmtgActiv != NaGuido::normal)
	  naGuido.tmtgMidiMessageList += midiNoteOn;
	else
	  {
	    // if oldTmtgActiv midiNoteOn was already written
	    if (naGuido.oldTmtgActiv == NaGuido::normal)
	      *(naGuido.currentMidiTrackChunk) += midiNoteOn;
	  }
      else
        {
          MidiMetaEvent *midiMetaEvent = new MidiMetaEvent ();
          midiMetaEvent->setDelay (midiNoteOn->delay ());
          midiMetaEvent->setMetaEventId (META_EVENT_TEXT);
          midiMetaEvent->setContent (MString ("tie"));
          *(naGuido.currentMidiTrackChunk) += midiMetaEvent;
          delete (midiNoteOn);
        }

      // write MidiNoteOff in NoteOffList
      MidiNoteOff *midiNoteOff = new MidiNoteOff;
      // multiply with delayFactor for \staccato etc.
      midiDelay = (unsigned long)((float)midiDelay * naGuido.delayFactor);
      // if trill, modent, turn, grace check for changing delay
      if ((naGuido.tmtgActiv != NaGuido::normal)&&(naGuido.tmtgPar > 0))
	midiDelay = ticksPerQuater () * 4 / naGuido.tmtgPar;
      midiNoteOff->setDelay (midiDelay);
      midiNoteOff->setChannel (naGuido.currentChannel);
      midiNoteOff->setKey (key + naGuido.oct * 12);
      // check intensity
      if (127 * naGuido.intensity * naGuido.intensityFactor < 127.0)
        midiNoteOn->setDynamic ((unsigned)(127 * naGuido.intensity * naGuido.intensityFactor));
      else
        midiNoteOn->setDynamic (127);

      // check for tenuto
      if (naGuido.tremActiv)
	{
	  unsigned long tremMidiDelay;
	  // if ten hadn't a option set default option and calculate 
	  // new delay for tenuto
	  if (naGuido.tremPar < 0)
	    tremMidiDelay = midiDelay / 4;
	  else
	    tremMidiDelay = midiDelay * durD / naGuido.tremPar;
	  // Write MidiNoteOn and MidiNoteOff
	  for (unsigned long l = 1; (l * tremMidiDelay) < midiDelay; l++)
            {
	      // Write MidiNoteOff with new delay
              MidiNoteOff *newMidiNoteOff = (MidiNoteOff*)midiNoteOff->copy ();
              newMidiNoteOff->setDelay (tremMidiDelay * l);
	      naGuido.midiNoteOffList += newMidiNoteOff;
              // Write MidiNoteOn one tick after MidiNotenOff
              MidiNoteOn *newMidiNoteOn = (MidiNoteOn*)midiNoteOn->copy ();
              newMidiNoteOn->setDelay (tremMidiDelay * l + 1);
	      naGuido.midiNoteOffList += newMidiNoteOn;
	    }
	}

      // if \tie is active store MidiNoteOff in tieMidiNoteOff and
      // else store it in MidiNoteOffList
      if (!naGuido.tie)
	// if trill, modent, turn, grace store MidiMessages in separat list
	if (naGuido.tmtgActiv != NaGuido::normal)
	  naGuido.tmtgMidiMessageList += midiNoteOff;
	else
	  naGuido.midiNoteOffList += midiNoteOff;
      else
        {
          if (naGuido.tieMidiNoteOff != 0)
            delete (naGuido.tieMidiNoteOff);
          naGuido.tieMidiNoteOff = midiNoteOff;
        }

      //reset oldTmtgActiv
      naGuido.oldTmtgActiv = NaGuido::normal;
    }
}

// ************************************************************ //
// append Tag                                                   //
// ************************************************************ //
class Tag
{
public:
  // Tag-variables
  MString name;
  int durPosN;
  int durPosD;
  // variables to save data for tag
  float f;
  long l;

  // Initialize Class
  Tag ()
    {
      durPosN = 0;
      durPosD = 1;
      f = 0.0;
      l = 0;
    }
};

typedef deque<Tag> TagQueue;
TagQueue tagQueue;

// start a Tag
void na_tagStart(char* s, int no, long int tagno, int durPosN, int durPosD)
{
  DEBUG(*(naGuido.logFile) << "na_tagStart(s=" << s << ",no=" << (long int)no << ",tagno=" << tagno << ",durPosN=" << (long int)durPosN << ",durPosD=" << (long int)durPosD << ")\n";)

  // initialize Tag
  Tag tag;
  tag.name = s;
  tag.durPosN = durPosN;
  tag.durPosD = durPosD;
  tagQueue.push_front (tag);
}

// all Arguments are red
void na_tagAdd(long int nargs)
{
  DEBUG(*(naGuido.logFile) << "na_tagAdd(nargs=" << nargs << ")\n";)

  // if there are some MidiNoteOff try to write the messages
  writeMidiNoteOffList (tagQueue.front ().durPosN, tagQueue.front ().durPosD);

  if ((strcmp(tagQueue.front ().name, "\\i")==0)||
      (strcmp(tagQueue.front ().name, "\\intens")==0))
    // Intensity \intens<s[,r]>
    //           \i<s[,r]>
    {
      // Check options
      if (nargs < 1)
        {
          (*naGuido.logFile) << MString ("Missing option for '\\intens<s[,r]>'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // write Intensity as MetaEvent-Text
      MidiMetaEvent *midiMetaEvent = new MidiMetaEvent ();
      midiMetaEvent->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiMetaEvent->setMetaEventId (META_EVENT_TEXT);
      midiMetaEvent->setContent (MString (gd_getTagArgStr (1)));
      *(naGuido.currentMidiTrackChunk) += midiMetaEvent;
      // set intensity, if tag has a second float argument
      if ((nargs == 2) ? gd_getTagArgType(2) == 'f' : false)
        naGuido.intensity = gd_getTagArgFloat(2);
      else
        {
          // look in dynamicList for arg1
          DynamicList::iterator iter;
          for (iter = naGuido.dynamicList.begin (); iter != naGuido.dynamicList.end (); ++iter)
            // if fond set intensity
            if ((*iter).name == MString (gd_getTagArgStr (1)))
              naGuido.intensity = (*iter).intensity;
        }
    }
  else if ((strcmp(tagQueue.front ().name, "\\cresc")==0)||
           (strcmp(tagQueue.front ().name, "\\dim")==0))
  // Cescendo \cresc[<s[,r]>](notes)
  // Decrescendo \dim[<s[,r]>](notes)
    {
      // if cresc/dim hat par r
      if (nargs == 2)
        {
          // save last position in front of \\cresc
          if (naGuido.crescendoBeginIter != 0)
            delete (naGuido.crescendoBeginIter);
          naGuido.crescendoBeginIter = new MidiMessageList::const_iterator (naGuido.currentMidiTrackChunk->end ());
          --(*naGuido.crescendoBeginIter);
          // save new intensity
          tagQueue.front ().f = gd_getTagArgFloat (2);
        }
      else
        tagQueue.front ().f = -1.0;
    }
  else if ((strcmp(tagQueue.front ().name, "\\crescBegin")==0)||
           (strcmp(tagQueue.front ().name, "\\dimBegin")==0))
  //          \crescBegin
  //             \dimBegin
    {
          // save last position in front of \\...Begin
          if (naGuido.crescendoBeginIter != 0)
            delete (naGuido.crescendoBeginIter);
          naGuido.crescendoBeginIter = new MidiMessageList::const_iterator (naGuido.currentMidiTrackChunk->end ());
          --(*naGuido.crescendoBeginIter);
    }
  else if ((strcmp(tagQueue.front ().name, "\\crescEnd")==0)||
           (strcmp(tagQueue.front ().name, "\\dimEnd")==0))
  //          \crescEnd[<s[,r]>]
  //             \dimEnd[<s[,r]>]
    {
      // check for \...Begin
      if (naGuido.crescendoBeginIter == 0)
        {
          (*naGuido.logFile) << MString ("Missing '\\crescBegin' or '\\dimBegin'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // if cresc/dim hat par r
      if (nargs == 2)
        tagQueue.front ().f = gd_getTagArgFloat (2);
      else
        {
          tagQueue.front ().f = -1.0;
          delete (naGuido.crescendoBeginIter);
          naGuido.crescendoBeginIter = 0;
        }
    }
  else if (strcmp(tagQueue.front ().name, "\\instr")==0)
    // Instrument \instr<s[,"MIDI n"]>
    {
      // Check options
      if (nargs < 1)
        {
          (*naGuido.logFile) << MString ("Missing option for '\\instr<s[,\"MIDI n\"]>'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // write instrumentname as MetaEvent-InstrumentName
      MidiMetaEvent *midiMetaEvent = new MidiMetaEvent ();
      midiMetaEvent->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiMetaEvent->setMetaEventId (META_EVENT_INSTRUMENT_NAME);
      midiMetaEvent->setContent (MString (gd_getTagArgStr (1)));
      *(naGuido.currentMidiTrackChunk) += midiMetaEvent;
      // set instrument, if tag has a second string argument
      if ((nargs == 2) ? gd_getTagArgType(2) == 's' : false)
        {
          MString instrument (gd_getTagArgStr(2));
          unsigned char instr = instrument.subString (
                    instrument.indexOf("MIDI ") + 5).asUnsigned ();
          MidiProgramChange *midiProgramChange = new MidiProgramChange ();
          midiProgramChange->setDelay (naGuido.delayAddieren);
          naGuido.delayAddieren = 0;
          midiProgramChange->setChannel (naGuido.currentChannel);
          midiProgramChange->setProgram (instr);
          *(naGuido.currentMidiTrackChunk) += midiProgramChange;
        }
    }
  else if (strcmp(tagQueue.front ().name, "\\tempo")==0)
    // Tempo \tempo<s[,"x/y=n"]>
    {
      // Check options
      if (nargs < 1)
        {
          (*naGuido.logFile) << MString ("Missing option for '\\tempo<s[,\"x/y=n\"]>'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // write temponame as MetaEvent-Text
      MidiMetaEvent *midiMetaEvent = new MidiMetaEvent ();
      midiMetaEvent->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiMetaEvent->setMetaEventId (META_EVENT_TEXT);
      midiMetaEvent->setContent (MString (gd_getTagArgStr (1)));
      *(naGuido.currentMidiTrackChunk) += midiMetaEvent;
      // set Tempo, if tag has a second string argument
      if ((nargs == 2) ? gd_getTagArgType(2) == 's' : false)
        {
          short zaehler, nenner, anzahl;
          MString tempo (gd_getTagArgStr(2));
          zaehler = MString (tempo).remove (tempo.indexOf ('/')).asUnsigned ();
          nenner = MString (tempo).subString (tempo.indexOf ('/')+1).
                     remove (tempo.indexOf ('=')).asUnsigned ();
          anzahl = MString (tempo).subString (tempo.indexOf ('=')+1).asUnsigned ();
          if (naGuido.multipelTempo)
            // if multipelTempo is allowed, save new quaterPerMinute
            naGuido.currentQuaterPerMinute = 4 * zaehler * anzahl / nenner;
          else
            {
              // write MetaSetTempo
              naGuido.quaterPerMinute = 4 * zaehler * anzahl / nenner;
              MidiMetaSetTempo *midiMetaSetTempo = new MidiMetaSetTempo ();
              midiMetaSetTempo->setDelay (naGuido.delayAddieren);
              naGuido.delayAddieren = 0;
              // usecPerQuater = usecPerMin / (qaterPerMinute)
              midiMetaSetTempo->setTempo (60000000 / naGuido.quaterPerMinute);
              *(naGuido.currentMidiTrackChunk) += midiMetaSetTempo;
            }
       }
    }
  else if ((strcmp(tagQueue.front ().name, "\\accel")==0)||
           (strcmp(tagQueue.front ().name, "\\rit")==0))
  // accelerando \accel[<s[,"x/y=n"]>](notes)
  // ritardando \rit[<s[,"x/y=n"]>](notes)
    {
      // if accel/rit hat par "x/y=n"
      if (nargs == 2)
        {
          // save last position in front of \rit
          if (naGuido.riterdandoBeginIter != 0)
            delete (naGuido.riterdandoBeginIter);
          naGuido.riterdandoBeginIter = new MidiMessageList::const_iterator (naGuido.currentMidiTrackChunk->end ());
          --(*naGuido.riterdandoBeginIter);
          // save new tempo e.g. quaterPerMinutes
          short zaehler, nenner, anzahl;
          MString tempo (gd_getTagArgStr(2));
          zaehler = MString (tempo).remove (tempo.indexOf ('/')).asUnsigned ();
          nenner = MString (tempo).subString (tempo.indexOf ('/')+1).
                     remove (tempo.indexOf ('=')).asUnsigned ();
          anzahl = MString (tempo).subString (tempo.indexOf ('=')+1).asUnsigned ();
          tagQueue.front ().l = 4 * zaehler * anzahl / nenner;
        }
      else
        tagQueue.front ().l = -1;
    }
  else if ((strcmp(tagQueue.front ().name, "\\accelBegin")==0)||
           (strcmp(tagQueue.front ().name, "\\ritBegin")==0))
  //             \accelBegin
  //            \ritBegin
    {
          // save last position in front of \\...Begin
          if (naGuido.riterdandoBeginIter != 0)
            delete (naGuido.riterdandoBeginIter);
          naGuido.riterdandoBeginIter = new MidiMessageList::const_iterator (naGuido.currentMidiTrackChunk->end ());
          --(*naGuido.riterdandoBeginIter);
    }
  else if ((strcmp(tagQueue.front ().name, "\\accelEnd")==0)||
           (strcmp(tagQueue.front ().name, "\\ritEnd")==0))
  //             \accelEnd[<s[,"x/y=n"]>]
  //            \ritEnd[<s[,"x/y=n"]>]
    {
      // check for \...Begin
      if (naGuido.riterdandoBeginIter == 0)
        {
          (*naGuido.logFile) << MString ("Missing '\\accelBegin' or '\\ritBegin'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // if accel/rit hat par "x/y=n"
      if (nargs == 2)
        {
          // save new tempo e.g. quaterPerMinutes
          short zaehler, nenner, anzahl;
          MString tempo (gd_getTagArgStr(2));
          zaehler = MString (tempo).remove (tempo.indexOf ('/')).asUnsigned ();
          nenner = MString (tempo).subString (tempo.indexOf ('/')+1).
                     remove (tempo.indexOf ('=')).asUnsigned ();
          anzahl = MString (tempo).subString (tempo.indexOf ('=')+1).asUnsigned ();
          tagQueue.front ().l = 4 * zaehler * anzahl / nenner;
        }
      else
        {
          tagQueue.front ().l = -1;
          delete (naGuido.riterdandoBeginIter);
          naGuido.riterdandoBeginIter = 0;
        }
    }
  else if (strcmp(tagQueue.front ().name, "\\staff")==0)
    // \staff<id,channel,port>
    {
      // Check options
      if ((nargs < 1)||((nargs > 1)&&(nargs != 3)))
        {
          (*naGuido.logFile) << MString ("Missing option for '\\staff<id[,channel,port]>'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // it there are 3 options ...
      if (nargs == 3)
	{
	  // set current channel
	  naGuido.currentChannel = gd_getTagArgInt(2);
	  
	  // set instrument
	  unsigned short instr = gd_getTagArgInt(3);
	  MidiProgramChange *midiProgramChange = new MidiProgramChange ();
	  midiProgramChange->setDelay (naGuido.delayAddieren);
	  naGuido.delayAddieren = 0;
	  midiProgramChange->setChannel (naGuido.currentChannel);
	  midiProgramChange->setProgram (instr);
	  *(naGuido.currentMidiTrackChunk) += midiProgramChange;
	}
    }
  else if (strcmp(tagQueue.front ().name, "\\alter")==0)
    // microtonal alterations: \alter<f>(notes)
    {
      // Check options
      if (nargs < 1)
        {
          (*naGuido.logFile) << MString ("Missing option for '\\alter<f>'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      float alter;
      if (gd_getTagArgType(1) == 'f')
        alter = gd_getTagArgFloat (1);
      else
        alter = (float)gd_getTagArgInt (1);
      // check reange of alter
      if (fabs (alter) > 2)
        {
          (*naGuido.logFile) << MString ("Wrong range of option for '\\alter<f>'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // write MidiPitchBend
      MidiPitchBendChange *midiPitchBendChange = new MidiPitchBendChange ();
      midiPitchBendChange->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiPitchBendChange->setChannel (naGuido.currentChannel);
      tagQueue.front ().l = naGuido.currentAlter;
      naGuido.currentAlter = (unsigned long)((alter+2.0) * 0x3FFF/4.0);
      midiPitchBendChange->setPitchBend (naGuido.currentAlter);
      *(naGuido.currentMidiTrackChunk) += midiPitchBendChange;
    }
  else if (strcmp(tagQueue.front ().name, "\\tie")==0)
  // tie \tie(notes)
    {
      // marke that tie is aktiv
      naGuido.tie = true;
    }
  else if (strcmp(tagQueue.front ().name, "\\stacc")==0)
  // staccato \stacc(notes)
    {
      tagQueue.front ().f = naGuido.delayFactor;
      naGuido.delayFactor = naGuido.factorStaccato;
    }
  else if (strcmp(tagQueue.front ().name, "\\accent")==0)
  // accented \accent(notes)
    {
      tagQueue.front ().f = naGuido.intensityFactor;
      naGuido.intensityFactor = naGuido.factorAccent;
    }
  else if (strcmp(tagQueue.front ().name, "\\marcato")==0)
  // strongly accented \marcato(notes)
    {
      tagQueue.front ().f = naGuido.intensityFactor;
      naGuido.intensityFactor = naGuido.factorMarcato;
    }
  else if (strcmp(tagQueue.front ().name, "\\ten")==0)
  // tenuto \ten(notes)
    {
      tagQueue.front ().f = naGuido.delayFactor;
      naGuido.delayFactor = naGuido.factorTenuto;
    }
  else if ((strcmp(tagQueue.front ().name, "\\slur")==0)||
           (strcmp(tagQueue.front ().name, "\\sl")==0))
  // tenuto \slur(notes), \sl(notes)
    {
      tagQueue.front ().f = naGuido.delayFactor;
      naGuido.delayFactor = naGuido.factorSlur;
    }
  else if ((strcmp(tagQueue.front ().name, "\\trill")==0)||
           (strcmp(tagQueue.front ().name, "\\mord")==0)||
           (strcmp(tagQueue.front ().name, "\\turn")==0)||
           (strcmp(tagQueue.front ().name, "\\grace")==0))
  // trill \trill[<i>](chords)
  // mordent \mord[<i>](chords)
  // turn \turn[<i>](chords)
  // \grace[<i>](notes)
    {
      // store old tag
      tagQueue.front ().l = naGuido.tmtgActiv;
      tagQueue.front ().f = (float)naGuido.tmtgPar;
      // set new tag
      if (strcmp(tagQueue.front ().name, "\\trill")==0)
	// set trill
	naGuido.tmtgActiv = NaGuido::trill;
      else if (strcmp(tagQueue.front ().name, "\\mord")==0)
        // set mordant
 	naGuido.tmtgActiv = NaGuido::mord;
      else if (strcmp(tagQueue.front ().name, "\\turn")==0)
        // set turn
	naGuido.tmtgActiv = NaGuido::turn;
      else if (strcmp(tagQueue.front ().name, "\\grace")==0)
        // set grace
	naGuido.tmtgActiv = NaGuido::grace;
      // check for option
      if (nargs == 1)
        naGuido.tmtgPar = gd_getTagArgInt (1);
      else
        naGuido.tmtgPar = -1;
    }
  else if (strcmp(tagQueue.front ().name, "\\trem")==0)
  // tremolo \trem[<i>](notes)
    {
      // set temulo
      naGuido.tremActiv = true;
      // check for option
      if (nargs == 1)
        naGuido.tremPar = gd_getTagArgInt (1);
      else
        naGuido.tremPar = -1;
    }
  else if (strcmp(tagQueue.front ().name, "\\fermata")==0)
  // \fermata(notes)
    naGuido.fermataActiv = true;
  else if (strcmp(tagQueue.front ().name, "\\cue")==0)
  // \cue<s>(notes)
    naGuido.cueActiv = true;
  else if (strcmp(tagQueue.front ().name, "\\repeatBegin")==0)
  // repeat \repeatBegin[<i>]
    {
      // write a MetaEvent-Text to set position
      MidiMetaEvent *midiMetaEvent = new MidiMetaEvent ();
      midiMetaEvent->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiMetaEvent->setMetaEventId (META_EVENT_TEXT);
      midiMetaEvent->setContent (MString ("\\repearBegin"));
      *(naGuido.currentMidiTrackChunk) += midiMetaEvent;
      // get number of repeats
      if (nargs == 1)
        naGuido.repeatNumber = gd_getTagArgInt (1);
      else
        naGuido.repeatNumber = 2;
      // number of current repeat;
      naGuido.currentRepeatNumber = 1;
      // Reset Iterators
      if (naGuido.repeatBeginIter != 0)
        delete (naGuido.repeatBeginIter);
      if (naGuido.repeatEndIter != 0)
        delete (naGuido.repeatEndIter);
      naGuido.repeatEndIter = 0;
      // save last position in front of \repeatbegin
      naGuido.repeatBeginIter = new MidiMessageList::const_iterator (naGuido.currentMidiTrackChunk->end ());
      --(*naGuido.repeatBeginIter);
    }
  else if (strcmp(tagQueue.front ().name, "\\repeatEnd")==0)
  //        \repeatEnd
  //        \repeatEnd<i>(notes)
    {
      // check if \repeatBegin was called
      if (naGuido.repeatBeginIter == 0)
        {
          (*naGuido.logFile) << MString ("Missing '\\repeatBegin'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // write a MetaEvent-Text to set position
      MidiMetaEvent *midiMetaEvent = new MidiMetaEvent ();
      midiMetaEvent->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiMetaEvent->setMetaEventId (META_EVENT_TEXT);
      midiMetaEvent->setContent (MString ("\\repearEnd"));
      *(naGuido.currentMidiTrackChunk) += midiMetaEvent;
      // save end position
      if (naGuido.repeatEndIter == 0)
        {
          naGuido.repeatEndIter = new MidiMessageList::const_iterator (naGuido.currentMidiTrackChunk->end ());
          --(*naGuido.repeatEndIter);
        }
      // get number of end
      if (nargs == 1)
        {
          // check number of repeat
          if (gd_getTagArgInt (1) != naGuido.currentRepeatNumber)
            {
              long l = gd_getTagArgInt (1);
              // write reteat
              while (l > naGuido.currentRepeatNumber)
                {
                  naGuido.currentMidiTrackChunk->insert (naGuido.currentMidiTrackChunk->end (),
                                                         ++(*naGuido.repeatBeginIter),
                                                         *naGuido.repeatEndIter);
                  *naGuido.currentMidiTrackChunk += **naGuido.repeatEndIter;
                  ++(naGuido.currentRepeatNumber);
                  --(*naGuido.repeatBeginIter);
                }
            }
        }
      else
        // write repeat
        while (naGuido.repeatNumber > naGuido.currentRepeatNumber)
          {
            naGuido.currentMidiTrackChunk->insert (naGuido.currentMidiTrackChunk->end (),
                                                   ++(*naGuido.repeatBeginIter),
                                                   *naGuido.repeatEndIter);
            *naGuido.currentMidiTrackChunk += **naGuido.repeatEndIter;
            ++(naGuido.currentRepeatNumber);
            --(*naGuido.repeatBeginIter);
          }
    }
  else if (strcmp(tagQueue.front ().name, "\\meter")==0)
    // meter \meter<"n/m">
    {
      // Check options
      if (nargs < 1)
        {
          (*naGuido.logFile) << MString ("Missing option for '\\meter<\"n/m\">'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // write META_EVENT_TIME_SIGNATURE
      unsigned char data[4];
      MString meter (gd_getTagArgStr(1));
      data[0] = MString (meter).remove (meter.indexOf ('/')).asUnsigned ();
      data[1] = MString (meter).subString (meter.indexOf ('/')+1).asUnsigned ();
      data[2] = 24;
      data[3] = 8;
      MidiMetaEvent *midiMetaEvent = new MidiMetaEvent ();
      midiMetaEvent->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiMetaEvent->setMetaEventId (META_EVENT_TIME_SIGNATURE);
      midiMetaEvent->setContent (data, 4);
      *(naGuido.currentMidiTrackChunk) += midiMetaEvent;
    }
  else if (strcmp(tagQueue.front ().name, "\\key")==0)
  // key \key<s>|<i>
    {
      // Check options
      if (nargs < 1)
        {
          (*naGuido.logFile) << MString ("Missing option for '\\key<s>|<i>'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // Write Meta Key signature
      signed char data[2];

      if (gd_getTagArgType(1) == 's')
        {
          MString mode (gd_getTagArgStr (1));
          // a List of possible modes
          if (strcmp (mode, "C") == 0)   // C major
            {  data[0] = 0;  data[1] = 0;  }
          else if (strcmp (mode, "G") == 0) // G major
            {  data[0] = 1;  data[1] = 0;  }
          else if (strcmp (mode, "D") == 0) // D major
            {  data[0] = 2;  data[1] = 0;  }
          else if (strcmp (mode, "A") == 0) // A major
            {  data[0] = 3;  data[1] = 0;  }
          else if (strcmp (mode, "E") == 0) // E major
            {  data[0] = 4;  data[1] = 0;  }
          else if (strcmp (mode, "H") == 0) // H major
            {  data[0] = 5;  data[1] = 0;  }
          else if (strcmp (mode, "F#") == 0) // F#major
            {  data[0] = 6;  data[1] = 0;  }
          else if (strcmp (mode, "C#") == 0) // C# major
            {  data[0] = 7;  data[1] = 0;  }
          else if (strcmp (mode, "F") == 0) // F major
            {  data[0] = -1;  data[1] = 0;  }
          else if (strcmp (mode, "H&") == 0) // H& major
            {  data[0] = -2;  data[1] = 0;  }
          else if (strcmp (mode, "E&") == 0) // E& major
            {  data[0] = -3;  data[1] = 0;  }
          else if (strcmp (mode, "A&") == 0) // A& major
            {  data[0] = -4;  data[1] = 0;  }
          else if (strcmp (mode, "D&") == 0) // D& major
            {  data[0] = -5;  data[1] = 0;  }
          else if (strcmp (mode, "G&") == 0) // G& major
            {  data[0] = -6;  data[1] = 0;  }
          else if (strcmp (mode, "C&") == 0) // C& major
            {  data[0] = -7;  data[1] = 0;  }
          else if (strcmp (mode, "a") == 0) // a minor
            {  data[0] = 0;  data[1] = 1;  }
          else if (strcmp (mode, "e") == 0) // e minor
            {  data[0] = 1;  data[1] = 1;  }
          else if (strcmp (mode, "h") == 0) // h minor
            {  data[0] = 2;  data[1] = 1;  }
          else if (strcmp (mode, "f#") == 0) // f# minor
            {  data[0] = 3;  data[1] = 1;  }
          else if (strcmp (mode, "c#") == 0) // c# minor
            {  data[0] = 4;  data[1] = 1;  }
          else if (strcmp (mode, "g#") == 0) // minor
            {  data[0] = 5;  data[1] = 1;  }
          else if (strcmp (mode, "d#") == 0) //  minor
            {  data[0] = 6;  data[1] = 1;  }
          else if (strcmp (mode, "a#") == 0) //  minor
            {  data[0] = 7;  data[1] = 1;  }
          else if (strcmp (mode, "d") == 0) //  minor
            {  data[0] = -1;  data[1] = 1;  }
          else if (strcmp (mode, "g") == 0) //  minor
            {  data[0] = -2;  data[1] = 1;  }
          else if (strcmp (mode, "c") == 0) //  minor
            {  data[0] = -3;  data[1] = 1;  }
          else if (strcmp (mode, "f") == 0) //  minor
            {  data[0] = -4;  data[1] = 1;  }
          else if (strcmp (mode, "h&") == 0) //  minor
            {  data[0] = -5;  data[1] = 1;  }
          else if (strcmp (mode, "e&") == 0) //  minor
            {  data[0] = -6;  data[1] = 1;  }
          else if (strcmp (mode, "a&") == 0) //  minor
            {  data[0] = -7;  data[1] = 1;  }
        }
      else
        {
          // first par ist the number of shorps or flats
          data[0] = gd_getTagArgInt(1);
          data[1] = 0; // major
        }
      MidiMetaEvent *midiMetaEvent = new MidiMetaEvent ();
      midiMetaEvent->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiMetaEvent->setMetaEventId (META_EVENT_KEY_SIGNATURE);
      midiMetaEvent->setContent ((unsigned char*)data, 2);
      *(naGuido.currentMidiTrackChunk) += midiMetaEvent;
    }
  else if (strcmp(tagQueue.front ().name, "\\oct")==0)
  // transposition \oct<i>
  //               \oct<i>(notes)
    {
      // Check options
      if (nargs < 1)
        {
          (*naGuido.logFile) << MString ("Missing option for '\\oct<i>'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // set oct
      tagQueue.front ().l  = naGuido.oct;
      naGuido.oct = gd_getTagArgInt(1);
    }
  else if ((strcmp(tagQueue.front ().name, "\\text")==0)||
           (strcmp(tagQueue.front ().name, "\\t")==0))
  // Text \text<s>
  //      \t<s>
    {
      // Check options
      if (nargs < 1)
        {
          (*naGuido.logFile) << MString ("Missing option for '\\text<s>'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // write Text as MetaEvent-Lyrics
      MidiMetaEvent *midiMetaEvent = new MidiMetaEvent ();
      midiMetaEvent->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiMetaEvent->setMetaEventId (META_EVENT_LYRICS);
      midiMetaEvent->setContent (MString (gd_getTagArgStr (1)));
      *(naGuido.currentMidiTrackChunk) += midiMetaEvent;
    }
  else if (strcmp(tagQueue.front ().name, "\\title")==0)
  // title \title<s>
    {
      // Check options
      if (nargs < 1)
        {
          (*naGuido.logFile) << MString ("Missing option for '\\title<s>'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // write titel as MetaEvent-TrackName
      MidiMetaEvent *midiMetaEvent = new MidiMetaEvent ();
      midiMetaEvent->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiMetaEvent->setMetaEventId (META_EVENT_TRACK_NAME);
      midiMetaEvent->setContent (MString (gd_getTagArgStr (1)));
      *(naGuido.currentMidiTrackChunk) += midiMetaEvent;
    }
  else if (strcmp(tagQueue.front ().name, "\\composer")==0)
  // composer \composer<s>
    {
      // Check options
      if (nargs < 1)
        {
          (*naGuido.logFile) << MString ("Missing option for '\\composer<s>'\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // write Composer as MetaEvent-Copyright
      MidiMetaEvent *midiMetaEvent = new MidiMetaEvent ();
      midiMetaEvent->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiMetaEvent->setMetaEventId (META_EVENT_COPYRIGHT);
      midiMetaEvent->setContent (MString (gd_getTagArgStr (1)));
      *(naGuido.currentMidiTrackChunk) += midiMetaEvent;
    }
  else if ((strcmp(tagQueue.front ().name, "\\mark")==0)||
           (strcmp(tagQueue.front ().name, "\\label")==0))
  // \mark<s>
  // \label<s>[(notes)]
    {
      // Check options
      if (nargs < 1)
        {
          (*naGuido.logFile) << MString ("Missing option for '\\mark<s>' or '\\label<s>\n");
          throw (Exception (ERROR_NAGUIDO));
        }
      // write Composer as MetaEvent-Text
      MidiMetaEvent *midiMetaEvent = new MidiMetaEvent ();
      midiMetaEvent->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiMetaEvent->setMetaEventId (META_EVENT_TEXT);
      midiMetaEvent->setContent (MString (gd_getTagArgStr (1)));
      *(naGuido.currentMidiTrackChunk) += midiMetaEvent;
    }
  else if ((strcmp(tagQueue.front ().name, "\\staff")==0)||
           (strcmp(tagQueue.front ().name, "\\beam")==0)||
           (strcmp(tagQueue.front ().name, "\\bm")==0)||
           (strcmp(tagQueue.front ().name, "\\beamsAuto")==0)||
           (strcmp(tagQueue.front ().name, "\\beamsOff")==0)||
           (strcmp(tagQueue.front ().name, "\\clef")==0)||
           (strcmp(tagQueue.front ().name, "\\stemsAuto")==0)||
           (strcmp(tagQueue.front ().name, "\\stemsUp")==0)||
           (strcmp(tagQueue.front ().name, "\\stemsDown")==0)||
           (strcmp(tagQueue.front ().name, "\\bar")==0)||
           (strcmp(tagQueue.front ().name, "\\doubleBar")==0)||
           (strcmp(tagQueue.front ().name, "\\tactus")==0))
  // unsupported: \staff, \beam, \bm, \beamsAuto, \beamsOff, \clef
  //              \stemsAuto, \stemsUp, \stemsDown, \bar, \doubleBar, \tactus
    // print Info unsupported Tag
    *(naGuido.logFile) << MString ("Unsupported tag: ") << tagQueue.front ().name << MString ("\n");
  else
    // print waring unknown tag
    *(naGuido.logFile) << MString ("WARING unknown tag: ") << tagQueue.front ().name << MString ("\n");
}

// end of Tag reached
void na_tagEnd(long int tagno, int durPosN, int durPosD,
               int beforeDurPosN, int beforeDurPosD)
{
  DEBUG(*(naGuido.logFile) << "na_tagEnd(tagno=" << (long int)tagno << ",durPosN=" << (long int)durPosN << ",durPosD=" << (long int)durPosD << ",beforeDurPosN=" << (long int)beforeDurPosN << ",beforeDurPosD=" << (long int)beforeDurPosD << ")\n";)

  // if there are som MidiNoteOff try to write the messages
  writeMidiNoteOffList (durPosN, durPosD);

  // search current tag
  if ((strcmp(tagQueue.front ().name, "\\cresc")==0)||
      (strcmp(tagQueue.front ().name, "\\crescEnd")==0))
  // Cescendo \cresc[<s[,r]>](notes)
  //          \crescEnd[<s[,r]>]
    {
      // if cresc/dim hat par r
      if (tagQueue.front ().f >= 0)
        {
          MidiMessageList::const_iterator iter = *naGuido.crescendoBeginIter;
          // calculate whole delay for cresc/dim
          unsigned long wholeDelay;
          for (wholeDelay = 0; iter != naGuido.currentMidiTrackChunk->end (); ++iter)
            wholeDelay += (*iter)->delay ();

          // search for MidiNoteOn or MidiNoteOff Messages
          unsigned long currentDelay;
          iter = *naGuido.crescendoBeginIter;
          for (currentDelay = 0; iter != naGuido.currentMidiTrackChunk->end (); ++iter)
            {
              currentDelay += (*iter)->delay ();
              // calculate and set new intensity
              float newDyn = naGuido.intensity + (tagQueue.front ().f - naGuido.intensity) * currentDelay / wholeDelay;
              if ((*iter)->id () == 0x90)  // MidiNoteOn
               ((MidiNoteOn*)(*iter))->setDynamic ((unsigned)(127 * newDyn));
              if ((*iter)->id () == 0x80)  // MidiNoteOff
               ((MidiNoteOff*)(*iter))->setDynamic ((unsigned)(127 * newDyn));
            }
          // set new intensity
          naGuido.intensity = tagQueue.front ().f;
        }
      // reset crescendoBeginIter
      delete (naGuido.crescendoBeginIter);
      naGuido.crescendoBeginIter = 0;
    }
  else if ((strcmp(tagQueue.front ().name, "\\dim")==0)||
           (strcmp(tagQueue.front ().name, "\\dimEnd")==0))
  // Decrescendo \dim[<s[,r]>](notes)
  //             \dimEnd[<s[,r]>]
    {
      // if cresc/dim hat par r
      if (tagQueue.front ().f >= 0)
        {
          MidiMessageList::const_iterator iter = *naGuido.crescendoBeginIter;
          // calculate whole delay for cresc/dim
          unsigned long wholeDelay;
          for (wholeDelay = 0; iter != naGuido.currentMidiTrackChunk->end (); ++iter)
            wholeDelay += (*iter)->delay ();

          // search for MidiNoteOn or MidiNoteOff Messages
          unsigned long currentDelay;
          iter = *naGuido.crescendoBeginIter;
          for (currentDelay = 0; iter != naGuido.currentMidiTrackChunk->end (); ++iter)
            {
              currentDelay += (*iter)->delay ();
              // calculate and set new intensity
              float newDyn = naGuido.intensity - (naGuido.intensity - tagQueue.front ().f) * currentDelay / wholeDelay;
              if ((*iter)->id () == 0x90)  // MidiNoteOn
               ((MidiNoteOn*)(*iter))->setDynamic ((unsigned)(127 * newDyn));
              if ((*iter)->id () == 0x80)  // MidiNoteOff
               ((MidiNoteOff*)(*iter))->setDynamic ((unsigned)(127 * newDyn));
            }
          // set new intensity
          naGuido.intensity = tagQueue.front ().f;
        }
      // reset crescendoBeginIter
      delete (naGuido.crescendoBeginIter);
      naGuido.crescendoBeginIter = 0;
    }
  else if ((strcmp(tagQueue.front ().name, "\\accel")==0)||
           (strcmp(tagQueue.front ().name, "\\accelEnd")==0))
  // accelerando \accel[<s[,"x/y=n"]>](notes)
  //             \accelEnd[<s[,"x/y=n"]>]
    {
      // if accel/rit hat par "x/y=n"
      if (tagQueue.front ().l >= 0)
        {
          MidiMessageList::const_iterator iter = *naGuido.riterdandoBeginIter;
          // calculate whole delay for accel/rit, skip first MidiMessage
          unsigned long wholeDelay;
          for (wholeDelay = 0,++iter; iter != naGuido.currentMidiTrackChunk->end (); ++iter)
            wholeDelay += (*iter)->delay ();

          // search for MidiNoteOn or MidiNoteOff Messages,
          // skip first MidiMessage
          unsigned long currentDelay;
          iter = *naGuido.riterdandoBeginIter;
          for (currentDelay = 0,++iter; iter != naGuido.currentMidiTrackChunk->end (); ++iter)
            {
              currentDelay += (*iter)->delay ();
              // calculate and set new delay
              float newQuaterPerMinute = (float)naGuido.currentQuaterPerMinute +
                (float)(tagQueue.front ().l - naGuido.currentQuaterPerMinute) *
                (float)currentDelay / (float)wholeDelay;
              (*iter)->setDelay ((unsigned long)((*iter)->delay () *
                (float)naGuido.currentQuaterPerMinute / newQuaterPerMinute));
            }
          // set new tempo, write MetaSetTempo
          MidiMetaSetTempo *midiMetaSetTempo = new MidiMetaSetTempo ();
          midiMetaSetTempo->setDelay (naGuido.delayAddieren);
          naGuido.delayAddieren = 0;
          // usecPerQuater = usecPerMin / (qaterPerMinute)
          midiMetaSetTempo->setTempo (60000000 / tagQueue.front ().l);
          *(naGuido.currentMidiTrackChunk) += midiMetaSetTempo;
        }
      // reset riterdandoBeginIter
      delete (naGuido.riterdandoBeginIter);
      naGuido.riterdandoBeginIter = 0;
    }
  else if ((strcmp(tagQueue.front ().name, "\\rit")==0)||
           (strcmp(tagQueue.front ().name, "\\ritEnd")==0))
  // ritardando \rit[<s[,"x/y=n"]>](notes)
  //            \ritEnd[<s[,"x/y=n"]>]
    {
      // if accel/rit hat par "x/y=n"
      if (tagQueue.front ().l >= 0)
        {
          MidiMessageList::const_iterator iter = *naGuido.riterdandoBeginIter;
          // calculate whole delay for accel/rit, skip first MidiMessage
          unsigned long wholeDelay;
          for (wholeDelay = 0,++iter; iter != naGuido.currentMidiTrackChunk->end (); ++iter)
            wholeDelay += (*iter)->delay ();

          // search for MidiNoteOn or MidiNoteOff Messages,
          // skip first MidiMessage
          unsigned long currentDelay;
          iter = *naGuido.riterdandoBeginIter;
          for (currentDelay = 0,++iter; iter != naGuido.currentMidiTrackChunk->end (); ++iter)
            {
              currentDelay += (*iter)->delay ();
              // calculate and set new delay
              float newQuaterPerMinute = (float)naGuido.currentQuaterPerMinute -
                (float)(naGuido.currentQuaterPerMinute - tagQueue.front ().l) *
                (float)currentDelay / (float)wholeDelay;
              (*iter)->setDelay ((unsigned long)((*iter)->delay () *
                newQuaterPerMinute / (float)naGuido.currentQuaterPerMinute));
            }
          // set new tempo, write MetaSetTempo
          MidiMetaSetTempo *midiMetaSetTempo = new MidiMetaSetTempo ();
          midiMetaSetTempo->setDelay (naGuido.delayAddieren);
          naGuido.delayAddieren = 0;
          // usecPerQuater = usecPerMin / (qaterPerMinute)
          midiMetaSetTempo->setTempo (60000000 / tagQueue.front ().l);
          *(naGuido.currentMidiTrackChunk) += midiMetaSetTempo;
        }
      // reset riterdandoBeginIter
      delete (naGuido.riterdandoBeginIter);
      naGuido.riterdandoBeginIter = 0;
    }
  else if (strcmp(tagQueue.front ().name, "\\tie")==0)
  // tie \tie(notes)
    {
      naGuido.tie = false;
      // store tieMidiNoteOff in currentMidiTrackChunk with new delay
      naGuido.tieMidiNoteOff->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      *(naGuido.currentMidiTrackChunk) += naGuido.tieMidiNoteOff;
      naGuido.tieMidiNoteOff = 0;
    }
  else if ((strcmp(tagQueue.front ().name, "\\stacc")==0)||
           (strcmp(tagQueue.front ().name, "\\ten")==0)||
           (strcmp(tagQueue.front ().name, "\\slur")==0)||
           (strcmp(tagQueue.front ().name, "\\sl")==0))
  // staccato \stacc(notes)
  // tenuto \ten(notes)
  // slur \slur (notes), \sl (notes)
    naGuido.delayFactor = tagQueue.front ().f;
  else if ((strcmp(tagQueue.front ().name, "\\accent")==0)||
           (strcmp(tagQueue.front ().name, "\\marcato")==0))
  // accented \accent(notes)
  // strongly accented \marcato(notes)
    naGuido.intensityFactor = tagQueue.front ().f;
  else if (strcmp(tagQueue.front ().name, "\\fermata")==0)
  // \fermata(notes)
    naGuido.fermataActiv = false;
  else if (strcmp(tagQueue.front ().name, "\\cue")==0)
  // \cue<s>(notes)
    naGuido.cueActiv = false;
  else if ((strcmp(tagQueue.front ().name, "\\trill")==0)||
	   (strcmp(tagQueue.front ().name, "\\mord")==0)||
	   (strcmp(tagQueue.front ().name, "\\turn")==0)||
	   (strcmp(tagQueue.front ().name, "\\grace")==0))
  // trill \trill[<i>](chords)
  // mordent \mord[<i>](chords)
  // turn \turn[<i>](chords)
  // \grace[<i>](notes)
    {
      // reset Tag
      naGuido.oldTmtgActiv = naGuido.tmtgActiv;
      naGuido.oldTmtgPar = naGuido.tmtgPar;
      naGuido.tmtgActiv = (NaGuido::TmtgActiv)tagQueue.front ().l;
      naGuido.tmtgPar = (signed long)tagQueue.front ().f;
    }
  else if (strcmp(tagQueue.front ().name, "\\trem")==0)
  // tremolo \trem[<i>](notes)
    {
      // reset tremolo
      naGuido.tremActiv = false;
      naGuido.tremPar = -1;
    }
  // if durPos != tagQueue.front ().durPos has the tag (...)
  else if ((durPosN != tagQueue.front ().durPosN)||
        (durPosD != tagQueue.front ().durPosD))
    if (strcmp(tagQueue.front ().name, "\\oct")==0)
      {
        // reset oct
        naGuido.oct = tagQueue.front ().l;
      }
    else if (strcmp(tagQueue.front ().name, "\\alter")==0)
      // microtonal alterations: \alter<f>(notes)
      {
      // reset currentAlter
      naGuido.currentAlter = tagQueue.front ().l;
      // write MidiPitchBend
      MidiPitchBendChange *midiPitchBendChange = new MidiPitchBendChange ();
      midiPitchBendChange->setDelay (naGuido.delayAddieren);
      naGuido.delayAddieren = 0;
      midiPitchBendChange->setChannel (naGuido.currentChannel);
      midiPitchBendChange->setPitchBend (naGuido.currentAlter);
      *(naGuido.currentMidiTrackChunk) += midiPitchBendChange;
    }

  // delete current Tag;
  tagQueue.pop_front ();
}

// ***************************************************************** //
// Tag-functions which are not needed                                //
// ***************************************************************** //

// Tag Argument: long int
void na_tagIntArg(long int)
{  }

// Tag Argument: double
void na_tagFloatArg(double)
{  }


// Tag Argument: char*
void na_tagStrArg(char)
{  }

