/*******************************************************************************
* Copyright Johannes Martin Software Entwicklung (L.Martin@TeamOS2.de)         *
********************************************************************************
* FILE NAME: MidiTrackChunk.cpp                                                *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Class implementation of the class:                                         *
*     MidiTrackChunk                                                          *
*******************************************************************************/

#include "MidiTrackChunk.hpp"
#include "MidiNoteOff.hpp"
#include "MidiNoteOn.hpp"
#include "MidiKeyPressure.hpp"
#include "MidiControlChange.hpp"
#include "MidiProgramChange.hpp"
#include "MidiChannelPressure.hpp"
#include "MidiPitchBendChange.hpp"
#include "MidiTimeCode.hpp"
#include "MidiSongPositionPointer.hpp"
#include "MidiSongSelect.hpp"
#include "MidiTuneRequest.hpp"
#include "MidiEndOfExclusive.hpp"
#include "MidiRealTimeMessages.hpp"
#include "MidiMetaEvent.hpp"
#include "MidiSystemExclusive.hpp"

#include "MidiErrors.h"

// |= Operator
MidiTrackChunk& MidiTrackChunk::operator|= (const MidiTrackChunk& trackTemp)
{
  // wenn Track Liste keine Elemente hat, fertig
  if (trackTemp.empty ())
    return *this;
  // Wenn Vorhandene Liste keine Elemente hat einfach Temp liste zuweisen.
  if (empty ())
    {
      *this = trackTemp;
      return *this;
    }

  // Neue MessageListe, MessageZeiger und Iteratoren zu den alten Listen
  MidiMessageList midiMessageListNeu;
  MidiMessage *message, *messageTemp;
  iterator iter = begin ();
  const_iterator iterTemp = trackTemp.begin ();

  // Kopie von den Ersten messages erstellen, damit der delay verandert
  // werden kann um diese dann nacher in die neue Liste einzufgen
  message = (*iter)->copy ();
  messageTemp = (*iterTemp)->copy ();

  // Alle Message der listen durchgehen und sorteieren. Deren Delay
  // entsprechend setzen
  while ((iter != end ())&&(iterTemp != trackTemp.end ()))
    {
      // delay ausrechnen und in Liste einfgen
      // Cursor entsprechen weiterschalten und neues Element holen
      if (message->delay () > messageTemp->delay ())
        {
          message->setDelay (message->delay () - messageTemp->delay ());
          midiMessageListNeu += messageTemp;
          if (++iterTemp != trackTemp.end ())
            messageTemp = (*iterTemp)->copy ();
        }
      else
        {
          messageTemp->setDelay (messageTemp->delay () - message->delay ());
          midiMessageListNeu += message;
          if (++iter != end ())
            message = (*iter)->copy ();
        }
    }

  // restlichen Messages anhngen
  if (iter != end ())
    {
      midiMessageListNeu += message;
      midiMessageListNeu.insert (midiMessageListNeu.end (), ++iter, end ());
    }
  if (iterTemp != trackTemp.end ())
    {
      midiMessageListNeu += messageTemp;
      midiMessageListNeu.insert (midiMessageListNeu.end (), ++iterTemp, trackTemp.end ());
    }

  *(MidiMessageList*)this = midiMessageListNeu;

  return *this;
}

// bestimmt die Naechste MidiMessage und liefert einen zeiger auf
// die neu erzeigte Message zurueck
MidiMessage* MidiTrackChunk::bestimmeMidiMessage (IO& io)
{
  // Wenn nchstes Zeichen ein Statusbyte ist Byte aus io lesen
  // Statusbyte, wenn oberstes Bit gesetzt
  if (io.peek () & 0x80)
    io >> lastStatusByte;

  MidiChannelMessage *midiChannelMessage;
  switch (lastStatusByte & 0xF0)
    {
      case 0x80:
           midiChannelMessage = new MidiNoteOff;
           midiChannelMessage->setChannel (lastStatusByte & 0x0F);
           return midiChannelMessage;
      case 0x90:
           midiChannelMessage = new MidiNoteOn;
           midiChannelMessage->setChannel (lastStatusByte & 0x0F);
           return midiChannelMessage;
      case 0xA0:
           midiChannelMessage = new MidiKeyPressure;
           midiChannelMessage->setChannel (lastStatusByte & 0x0F);
           return midiChannelMessage;
      case 0xB0:
           midiChannelMessage = new MidiControlChange;
           midiChannelMessage->setChannel (lastStatusByte & 0x0F);
           return midiChannelMessage;
      case 0xC0:
           midiChannelMessage = new MidiProgramChange;
           midiChannelMessage->setChannel (lastStatusByte & 0x0F);
           return midiChannelMessage;
      case 0xD0:
           midiChannelMessage = new MidiChannelPressure;
           midiChannelMessage->setChannel (lastStatusByte & 0x0F);
           return midiChannelMessage;
      case 0xE0:
           midiChannelMessage = new MidiPitchBendChange;
           midiChannelMessage->setChannel (lastStatusByte & 0x0F);
           return midiChannelMessage;
      case 0xF0:
           switch (lastStatusByte)
             {
               case 0xF0:
                    return new MidiSystemExclusive ();
               case 0xF1:
                    return new MidiTimeCode ();
               case 0xF2:
                    return new MidiSongPositionPointer ();
               case 0xF3:
                    return new MidiSongSelect ();
// 0xF4, 0xF5 ?
               case 0xF6:
                    return new MidiTuneRequest ();
               case 0xF7:
                    return new MidiEndOfExclusive ();
               case 0xF8:
                    return new MidiTimingClock ();
// 0xF9 ?
               case 0xFA:
                    return new MidiStart ();
               case 0xFB:
                    return new MidiContinue ();
               case 0xFC:
                    return new MidiStop ();
// 0xFD ?
               case 0xFE:
                    return new MidiActiveSensing ();
               case 0xFF:
                    return bestimmeMidiMetaEvent (io);
             }
    }

  throw (Exception (ERROR_MIDI_NO_MESSAGE));
}

// Tracklnge erfragen in msec
unsigned long MidiTrackChunk::length (const unsigned short ticksPerQuater) const
{
  unsigned long mSec = 0;
  unsigned long uSecPerMidiTick = 500000 / ticksPerQuater; // uSecPerQuater / ticksPerQuater

  // Iterator anlegen und an anfang setzen
  const_iterator iter = begin ();

  while (iter != end ())
    {
      // Lnge aktualisieren setzen
      mSec += ((*iter)->delay () * uSecPerMidiTick) / 1000;

      // Wenn MetaSetTempo uSecPerMidiTick aktualisieren
      if ((*iter)->id () == 0xFF)
        if (((MidiMetaEvent*)*iter)->metaEventId () == 0x51)
          {
            uSecPerMidiTick = ((MidiMetaSetTempo*)*iter)->tempo () / ticksPerQuater;
          }
      ++iter;
    }

  return (mSec);
}

// Tracklnge erfragen in ticks
unsigned long MidiTrackChunk::length () const
{
  unsigned long ticks = 0;

  // Iterator anlegen und an anfang setzen
  const_iterator iter = begin ();

  while (iter != end ())
    {
      // Lnge aktualisieren setzen
      ticks += (*iter)->delay ();
      ++iter;
    }

  return (ticks);
}

// Titel erfragen
MString MidiTrackChunk::titel () const
{
  // Iterator anlegen und an anfang setzen
  const_iterator iter = begin ();

  // Nach Spurname suchen
  while (iter != end ())
    {
      if ((*iter)->id () == 0xFF)
        if (((const MidiMetaEvent*)*iter)->metaEventId () == 0x03) // Spurname
          return ((const MidiMetaEvent*)*iter)->content ();

      ++iter;
    }

  return "";
}

// Copyright erfragen
MString MidiTrackChunk::copyright () const
{
  // Iterator anlegen und an anfang setzen
  const_iterator iter = begin ();

  // Nach Copyright suchen
  while (iter != end ())
    {
      if ((*iter)->id () == 0xFF)
        if (((const MidiMetaEvent*)*iter)->metaEventId () == 0x02) // Copyright
          return ((const MidiMetaEvent*)*iter)->content ();

      ++iter;
    }

  return "";
}

// sichern
FileIO& operator<< (FileIO& fileIO, const MidiTrackChunk& midiTrackChunk)
{
  fileIO << (unsigned char)'M' << (unsigned char)'T'
           << (unsigned char)'r' << (unsigned char)'k';

  // fr lnge Platzhalter einsetzen
  long int laengenPosition = fileIO.tell ();
  fileIO << (unsigned long) 0;

  // Messages Schreiben
  // Iterator anlegen und an anfang setzen
  MidiMessageList::const_iterator iter = midiTrackChunk.begin ();

  while (iter != midiTrackChunk.end ())
    {
       fileIO << (*iter)->delay () << **iter;
       ++iter;
    }

  // Wenn Track nicht mit MidiMetaEndOfTrack abgeschlossen ist
  // MidiMetaEndOfTrack schreiben
  if ((iter == midiTrackChunk.begin ()) ? true :
      ((*(--iter))->id () != MidiMetaEvent ().id () ? true :
       (((MidiMetaEvent*)*iter)->metaEventId () != MidiMetaEndOfTrack ().metaEventId ())))
    fileIO << MidiDelay () << MidiMetaEndOfTrack ();

  // Lnge schreiben
  long int endPosition = fileIO.tell ();
  fileIO.seek (laengenPosition);
  fileIO << (unsigned long)(endPosition - laengenPosition - sizeof (unsigned long));
  fileIO.seek (endPosition);

  return fileIO;
}

// laden
FileIO& operator>> (FileIO& fileIO, MidiTrackChunk& midiTrackChunk)
{
  unsigned char c, c2, c3, c4;

  // wenn erstes Zeichen nicht 'M', war es eventuell noch zum auffllen des
  // Tracks auf Wortlnge gedacht
  fileIO >> c;
  if (c != 'M')
    fileIO >> c;
  // Kennung berprfen
  fileIO >> c2 >> c3 >> c4;
  if ((c != 'M')||(c2 != 'T')||(c3 != 'r')||(c4 != 'k'))
    throw (Exception (ERROR_MIDI_TRACK_CHUNK));

  // Lnge einlesen
  unsigned long laenge;
  fileIO >> laenge;
  long int beginPosition = fileIO.tell ();

  // Alten Messages lschen
  midiTrackChunk.erase (midiTrackChunk.begin (), midiTrackChunk.end ());

  // Messages einlesen
  while (beginPosition + laenge > (unsigned long)fileIO.tell ())
    {
      MidiDelay midiDelay;
      fileIO >> midiDelay;

      MidiMessage *midiMessage = midiTrackChunk.bestimmeMidiMessage (fileIO);
      fileIO >> *midiMessage;
      midiMessage->setDelay (midiDelay);

      midiTrackChunk += midiMessage;
    }

  // Wenn letztes Element ein MidiMetaEndOfTrack dann loeschen
  MidiMessageList::iterator iter = midiTrackChunk.end ();
  --iter;
  if ((*iter)->id () != MidiMetaEvent ().id () ? false :
      (((MidiMetaEvent*)*iter)->metaEventId () == MidiMetaEndOfTrack ().metaEventId ()))
    midiTrackChunk.erase (iter);

  return fileIO;
}


